import json
import requests
import boto3
from requests_aws4auth import AWS4Auth
from urllib.parse import urlparse
import cfnresponse
import time

def lambda_handler(event, context):
    print("Event:", event)
    model_id_list = event['ModelIdList']

    if len(model_id_list) == 0:
        return {'statusCode': 200,
        'body': json.dumps('ModelIdList is empty. Job completed successfully with no actions taken.')}

    resource_properties = event['ResourceProperties']
    host = get_aos_endpoint(resource_properties)
    aos_role_arn = get_aos_role_arn(resource_properties)
    
    awsauth = create_aws_auth(context, host)
    
    if not host.endswith('/'):
        host += '/'
        
    print("host:" , host)
    
    ######################################################
    # get data node ids 
    ######################################################
    
    nodel_url = host + '_nodes/data:true/stats'
    
    node_response = requests.get(url=nodel_url, auth=awsauth, headers={"Content-Type": "application/json"})
    
    data = node_response.json()
    data_nodes = list(data["nodes"].keys())
    print("data_nodes" + str(data_nodes))

    ######################################################
    # check models in model_id_list
    ######################################################

    profile_url = host + '_plugins/_ml/profile/models/'
    undeployed_model_ids = set()
    for model_id in model_id_list:
        url = profile_url + model_id
        profile_response = requests.get(url, auth=awsauth)
        
        data = profile_response.json()
        if data:
            for node_id, models in data["nodes"].items(): 
                #skipping master nodes 
                if node_id in data_nodes:
                    if "models" in models:
                        for model_id, model_info in models["models"].items():
                            if "model_state" in model_info.keys():
                                model_state = model_info["model_state"]
                                if model_state == "UNDEPLOYED": 
                                    undeployed_model_ids.add(model_id)
                            # there is no model_state in the response ,   
                            # but is a data node, should add to undeploy list
                            else: 
                                undeployed_model_ids.add(model_id)
                    else:
                        undeployed_model_ids.add(model_id)

        # the model is not deployed to any nodes
        else:
            undeployed_model_ids.add(model_id)
            print("model is not deploy in any nodes:" + model_id)
            
    # deploying models
    if len(undeployed_model_ids) > 0:
        print("Detected undeployed model id:", undeployed_model_ids)
        for model_id in undeployed_model_ids:
            deploy_url = host + '_plugins/_ml/models/'+ model_id +'/_deploy'
            
            deploy_response = requests.post(deploy_url, auth=awsauth, headers={"Content-Type": "application/json"})
            data = deploy_response.json()
            if "task_id" in data:
                task_id = data["task_id"]
                
                get_task_url = host + '_plugins/_ml/tasks/'+ task_id
            
                task_response = requests.get(get_task_url, auth=awsauth, headers={"Content-Type": "application/json"})
                status = task_response.json()["state"]
                
                print("deploy model task ids: ", task_id)
                print("deploy model task status: ", status)
                
                while status in ("CREATED","RUNNING"):
                    time.sleep(5)
                    task_response = requests.get(get_task_url, auth=awsauth, headers={"Content-Type": "application/json"})
                    status = task_response.json()["state"]
                    print(f"poll deploy model task status: {status}")
            else:
                print(f"Error happens when deploy model {model_id}: {data}")
                
        print("Completed model deployment for model id: ", undeployed_model_ids)
        
            
        return {'statusCode': 200,
            'body': json.dumps('Auto-deployment succeeded with models: ' + ", ".join(undeployed_model_ids))}
    else: 
        return {'statusCode': 200, 'body': json.dumps('No Auto-deployment needed.' )}
            


def create_aws_auth(context, host):
    region = context.invoked_function_arn.split(':')[3]
    service = 'es'
    session = boto3.Session()
    credentials = session.get_credentials()
    awsauth = AWS4Auth(credentials.access_key, credentials.secret_key, region, service, session_token=credentials.token)
    return awsauth

def get_aos_endpoint(resource_properties):
    if "AOSEndpoint" not in resource_properties:
        raise Exception("AOSEndpoint not provided from resource properties")
    return resource_properties["AOSEndpoint"]

def get_aos_role_arn(resource_properties):
    if "AOSRoleArn" not in resource_properties:
        raise Exception("AOSRoleArn not provided from resource properties")
    return resource_properties["AOSRoleArn"]
